@@ -29,9 +29,7 @@ static void ftp_got_banner(struct connection *, struct read_buffer *);
 static void ftp_login(struct connection *);
 static void ftp_logged(struct connection *);
 static void ftp_sent_passwd(struct connection *);
-static void ftp_got_info(struct connection *, struct read_buffer *);
 static void ftp_got_user_info(struct connection *, struct read_buffer *);
-static void ftp_dummy_info(struct connection *, struct read_buffer *);
 static void ftp_pass_info(struct connection *, struct read_buffer *);
 static void ftp_send_retr_req(struct connection *, int, int);
 static struct ftp_connection_info *add_file_cmd_to_str(struct connection *, int);
@@ -51,9 +49,9 @@ static int get_ftp_response(struct connection *c, struct read_buffer *rb, int pa
 	set_connection_timeout(c);
 	again:
 	for (l = 0; l < rb->len; l++) if (rb->data[l] == 10) {
-		unsigned char *e;
-		long k = strtoul(cast_const_char rb->data, (char **)(void *)&e, 10);
-		if (e != rb->data + 3 || k < 100 || k >= 1000) return -1;
+		char *e;
+		long k = strtoul(cast_const_char rb->data, &e, 10);
+		if (cast_uchar e != rb->data + 3 || k < 100 || k >= 1000) return -1;
 		if (*e == '-') {
 			int i;
 			for (i = 0; i < rb->len - 5; i++) {
@@ -100,7 +98,7 @@ void ftp_func(struct connection *c)
 			abort_connection(c);
 			return;
 		}
-		make_connection(c, p, &c->sock1, ftp_options.fast_ftp ? ftp_login : ftp_get_banner);
+		make_connection(c, p, &c->sock1, ftp_get_banner);
 	} else {
 		ftp_send_retr_req(c, S_SENT, we_are_in_root);
 	}
@@ -148,19 +146,7 @@ static void ftp_login(struct connection *c)
 	if ((u = get_user_name(c->url)) && *u) add_to_str(&login, &logl, u);
 	else add_to_str(&login, &logl, cast_uchar "anonymous");
 	if (u) mem_free(u);
-	if (ftp_options.fast_ftp) {
-		struct ftp_connection_info *fi;
-		add_to_str(&login, &logl, cast_uchar "\r\nPASS ");
-		u = get_ftp_password(c);
-		add_to_str(&login, &logl, u);
-		mem_free(u);
-		add_to_str(&login, &logl, cast_uchar "\r\n");
-		if (!(fi = add_file_cmd_to_str(c, 0))) {
-			mem_free(login);
-			return;
-		}
-		add_to_str(&login, &logl, fi->cmdbuf);
-	} else add_to_str(&login, &logl, cast_uchar "\r\n");
+	add_to_str(&login, &logl, cast_uchar "\r\n");
 	write_to_socket(c, c->sock1, login, logl, ftp_logged);
 	mem_free(login);
 	setcstate(c, S_SENT);
@@ -170,19 +156,6 @@ static void ftp_logged(struct connection *c)
 {
 	struct read_buffer *rb;
 	if (!(rb = alloc_read_buffer(c))) return;
-	if (!ftp_options.fast_ftp) {
-		ftp_got_user_info(c, rb);
-		return;
-	}
-	read_from_socket(c, c->sock1, rb, ftp_got_info);
-}
-
-static void ftp_got_info(struct connection *c, struct read_buffer *rb)
-{
-	int g = get_ftp_response(c, rb, 0);
-	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
-	if (!g) { read_from_socket(c, c->sock1, rb, ftp_got_info); return; }
-	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); retry_connection(c); return; }
 	ftp_got_user_info(c, rb);
 }
 
@@ -194,35 +167,23 @@ static void ftp_got_user_info(struct connection *c, struct read_buffer *rb)
 	if (g >= 530 && g < 540) { setcstate(c, S_FTP_LOGIN); retry_connection(c); return; }
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); retry_connection(c); return; }
 	if (g >= 200 && g < 300) {
-		if (ftp_options.fast_ftp) ftp_dummy_info(c, rb);
-		else ftp_send_retr_req(c, S_GETH, 0);
+		ftp_send_retr_req(c, S_GETH, 0);
 	} else {
-		if (ftp_options.fast_ftp) ftp_pass_info(c, rb);
-		else {
-			unsigned char *login;
-			unsigned char *u;
-			int logl = 0;
-			login = init_str();
-			add_to_str(&login, &logl, cast_uchar "PASS ");
-			u = get_ftp_password(c);
-			add_to_str(&login, &logl, u);
-			mem_free(u);
-			add_to_str(&login, &logl, cast_uchar "\r\n");
-			write_to_socket(c, c->sock1, login, logl, ftp_sent_passwd);
-			mem_free(login);
-			setcstate(c, S_LOGIN);
-		}
+		unsigned char *login;
+		unsigned char *u;
+		int logl = 0;
+		login = init_str();
+		add_to_str(&login, &logl, cast_uchar "PASS ");
+		u = get_ftp_password(c);
+		add_to_str(&login, &logl, u);
+		mem_free(u);
+		add_to_str(&login, &logl, cast_uchar "\r\n");
+		write_to_socket(c, c->sock1, login, logl, ftp_sent_passwd);
+		mem_free(login);
+		setcstate(c, S_LOGIN);
 	}
 }
 
-static void ftp_dummy_info(struct connection *c, struct read_buffer *rb)
-{
-	int g = get_ftp_response(c, rb, 0);
-	if (g == -1) { setcstate(c, S_FTP_ERROR); abort_connection(c); return; }
-	if (!g) { read_from_socket(c, c->sock1, rb, ftp_dummy_info); return; }
-	ftp_retr_file(c, rb);
-}
-
 static void ftp_sent_passwd(struct connection *c)
 {
 	struct read_buffer *rb;
@@ -237,8 +198,7 @@ static void ftp_pass_info(struct connection *c, struct read_buffer *rb)
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_pass_info); setcstate(c, S_LOGIN); return; }
 	if (g >= 530 && g < 540) { setcstate(c, S_FTP_LOGIN); abort_connection(c); return; }
 	if (g >= 400) { setcstate(c, S_FTP_UNAVAIL); abort_connection(c); return; }
-	if (ftp_options.fast_ftp) ftp_retr_file(c, rb);
-	else ftp_send_retr_req(c, S_GETH, 0);
+	ftp_send_retr_req(c, S_GETH, 0);
 }
 
 static void add_port_pasv(unsigned char **s, int *l, struct ftp_connection_info *inf, unsigned char *port_string)
@@ -312,7 +272,7 @@ static struct ftp_connection_info *add_file_cmd_to_str(struct connection *c, int
 				sprintf(cast_char port_string, "%d,%d,%d,%d,%d,%d", pc[0], pc[1], pc[2], pc[3], pc[4], pc[5]);
 		}
 		if (strlen(cast_const_char port_string) >= sizeof(port_string))
-			internal("buffer overflow in get_pasv_socket_ipv6: %d > %d", (int)strlen(cast_const_char port_string), (int)sizeof(port_string));
+			internal_error("buffer overflow in get_pasv_socket_ipv6: %d > %d", (int)strlen(cast_const_char port_string), (int)sizeof(port_string));
 	}
 #ifdef HAVE_IPTOS
 	if (ftp_options.set_tos) {
@@ -382,7 +342,7 @@ static void ftp_send_retr_req(struct connection *c, int state, int we_are_in_roo
 	struct ftp_connection_info *fi;
 	unsigned char *login;
 	int logl = 0;
-	set_connection_timeout(c);
+	set_connection_timeout_keepal(c);
 	login = init_str();
 	if (!c->info && !(fi = add_file_cmd_to_str(c, we_are_in_root))) {
 		mem_free(login);
@@ -390,7 +350,7 @@ static void ftp_send_retr_req(struct connection *c, int state, int we_are_in_roo
 	} else {
 		fi = c->info;
 	}
-	if (ftp_options.fast_ftp) {
+	if (0) {
 		a:add_to_str(&login, &logl, fi->cmdbuf);
 	} else {
 		unsigned char *nl = cast_uchar strchr(cast_const_char fi->cmdbuf, '\n');
@@ -417,10 +377,8 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 	struct ftp_connection_info *inf = c->info;
 	if (0) {
 		rep:
-		if (!ftp_options.fast_ftp) {
-			ftp_send_retr_req(c, S_GETH, inf->we_are_in_root);
-			return;
-		}
+		ftp_send_retr_req(c, S_GETH, inf->we_are_in_root);
+		return;
 	}
 	if (inf->pending_commands > 1) {
 		unsigned char pc[6] = { 0, 0, 0, 0, 0, 0 }; /* against warning */
@@ -504,7 +462,7 @@ static void ftp_retr_file(struct connection *c, struct read_buffer *rb)
 				}
 				goto rep;
 		}
-		internal("WHAT???");
+		internal_error("WHAT???");
 	}
 	g = get_ftp_response(c, rb, 2);
 	if (!g) { read_from_socket(c, c->sock1, rb, ftp_retr_file); setcstate(c, S_GETH); return; }
@@ -556,7 +514,6 @@ static void ftp_got_final_response(struct connection *c, struct read_buffer *rb)
 		}
 		if (c->cache->redirect) mem_free(c->cache->redirect);
 		c->cache->redirect = stracpy(c->url);
-		c->cache->redirect_get = 1;
 		add_to_strn(&c->cache->redirect, cast_uchar "/");
 		c->cache->incomplete = 0;
 		/*setcstate(c, S_FTP_NO_FILE);*/
@@ -619,17 +576,12 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 	/*add_to_str(&str, &sl, cast_uchar "   ");*/
 
 	dir = 0;
-	if (!*d || *d == PARSE_MODE_VMS) for (ee = 0; ee + 1 < p && !WHITECHAR(buf[ee]); ee++) {
+	if (!*d || *d == PARSE_MODE_VMS) for (ee = 0; p - ee > 1 && !WHITECHAR(buf[ee]); ee++) {
 		if (buf[ee] == ';' && buf[ee + 1] >= '1' && buf[ee + 1] <= '9') {
 			if (!ee) goto skip_vms;
 			if (ee >= 4 && buf[ee - 4] == '.' && buf[ee - 3] == 'D' && buf[ee - 2] == 'I' && buf[ee - 1] == 'R') {
 				if (ee == 4) goto raw;
 				dir = 1, ee -= 4;
-			} else {
-				/*if (buf[ee + 1] == '1' && (ee + 2 == p || buf[ee + 2] < '0' || buf[ee + 2] > '9')) goto no_version;
-				ee += 2;
-				while (ee < p && buf[ee] >= '0' && buf[ee] < '9') ee++;
-				no_version:;*/
 			}
 			if (*d != PARSE_MODE_VMS && !root) {
 				add_to_str(&str, &sl, cast_uchar "<a href=\"../\">..</a>\n");
@@ -700,7 +652,7 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 			    buf[pp - 2] >= '0' && buf[pp - 2] <= '5' &&
 			    buf[pp - 1] >= '0' && buf[pp - 1] <= '9') {
 				ppos = pp;
-				if (pp + 2 < p && buf[pp + 1] == ' ' && buf[pp + 2] != ' ')
+				if (p - pp > 2 && buf[pp + 1] == ' ' && buf[pp + 2] != ' ')
 					ppos++;
 			}
 		}
@@ -709,11 +661,11 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 			goto done;
 		}
 
-		for (pp = 0; pp + 5 <= p; pp++)
+		for (pp = 0; p - pp >= 5; pp++)
 			if (!casecmp(&buf[pp], cast_uchar "<DIR>", 5)) {
 				pp += 4;
-				while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
-				if (pp + 1 < p) goto done;
+				while (p - pp > 1 && WHITECHAR(buf[pp + 1])) pp++;
+				if (p - pp > 1) goto done;
 			}
 
 		bn = -1;
@@ -736,8 +688,8 @@ static int ftp_process_dirlist(struct cache_entry *ce, off_t *pos, int *d, unsig
 		}
 		if (bn >= 0) {
 			pp = bp + bn;
-			while (pp + 1 < p && WHITECHAR(buf[pp + 1])) pp++;
-			if (pp + 1 < p) goto done;
+			while (p - pp > 1 && WHITECHAR(buf[pp + 1])) pp++;
+			if (p - pp > 1) goto done;
 		}
 
 		for (pp = p - 1; pp >= 0; pp--) if (!WHITECHAR(buf[pp])) break;
