@@ -99,35 +99,26 @@ static void separate_fg_bg(int *fgr, int *fgg, int *fgb
 	}
 }
 
-static unsigned char *make_html_font_name(int attr)
-{
-	unsigned char *str;
-	int len;
-
-	str=init_str();len=0;
-	add_to_str(&str, &len, cast_uchar G_HTML_DEFAULT_FAMILY);
-	add_to_str(&str, &len, attr & AT_BOLD ? cast_uchar "-bold" : cast_uchar "-medium");
-	add_to_str(&str, &len, attr & AT_ITALIC ? cast_uchar "-italic-serif" : cast_uchar "-roman-serif");
-	add_to_str(&str, &len, attr & AT_FIXED ? cast_uchar "-mono" : cast_uchar "-vari");
-	return str;
-}
-
 static struct style *get_style_by_ta(struct text_attrib *ta)
 {
-	int fg_r,fg_g,fg_b; /* sRGB 0-255 values */
+	int fg_r, fg_g, fg_b; /* sRGB 0-255 values */
 	int fs = get_real_font_size(ta->fontsize);
-	struct style*stl;
-	unsigned char *fontname;
+	struct style *stl;
+	unsigned fflags;
+
+	fg_r = ta->fg.r;
+	fg_g = ta->fg.g;
+	fg_b = ta->fg.b;
+	separate_fg_bg(&fg_r, &fg_g, &fg_b, ta->bg.r, ta->bg.g, ta->bg.b);
+
+	fflags = 0;
+	if (ta->attr & AT_UNDERLINE) fflags |= FF_UNDERLINE;
+	if (ta->attr & AT_BOLD) fflags |= FF_BOLD;
+	if (ta->attr & AT_ITALIC) fflags |= FF_ITALIC;
+	if (ta->attr & AT_FIXED) fflags |= FF_MONOSPACED;
 
-	fg_r=ta->fg.r;
-	fg_g=ta->fg.g;
-	fg_b=ta->fg.b;
-	separate_fg_bg(&fg_r,&fg_g,&fg_b,ta->bg.r,ta->bg.g,ta->bg.b);
 	stl = g_get_style((fg_r << 16) + (fg_g << 8) + fg_b, (ta->bg.r << 16) +
-			(ta->bg.g << 8) + ta->bg.b, fs,
-			fontname=make_html_font_name(ta->attr),
-			ta->attr & AT_UNDERLINE ? FF_UNDERLINE : 0);
-	mem_free(fontname);
+			(ta->bg.g << 8) + ta->bg.b, fs, fflags);
 	return stl;
 }
 
@@ -147,7 +138,7 @@ void flush_pending_line_to_obj(struct g_part *p, int minheight)
 			struct g_object_text *got = get_struct(go, struct g_object_text, goti.go);
 			int l = (int)strlen(cast_const_char got->text);
 			while (l && got->text[l - 1] == ' ') got->text[--l] = 0, got->goti.go.xw -= g_char_width(got->style, ' ');
-			if (got->goti.go.xw < 0) internal("xw(%d) < 0", got->goti.go.xw);
+			if (got->goti.go.xw < 0) internal_error("xw(%d) < 0", got->goti.go.xw);
 		}
 		if (!go->xw) continue;
 		break;
@@ -209,7 +200,7 @@ void add_object_to_line(struct g_part *pp, struct g_object_line **lp, struct g_o
 {
 	struct g_object_line *l;
 	if (go && (go->xw < 0 || go->yw < 0)) {
-		internal("object has negative size: %d,%d", go->xw, go->yw);
+		internal_error("object has negative size: %d,%d", go->xw, go->yw);
 		return;
 	}
 	if (!*lp) {
@@ -274,7 +265,7 @@ static void split_line_object(struct g_part *p, struct g_object *text_go, unsign
 	}
 #ifdef DEBUG
 	if (ptr < text_t->text || ptr >= text_t->text + strlen(cast_const_char text_t->text))
-		internal("split_line_object: split point (%p) pointing out of object (%p,%lx)", ptr, text_t->text, (unsigned long)strlen(cast_const_char text_t->text));
+		internal_error("split_line_object: split point (%p) pointing out of object (%p,%lx)", ptr, text_t->text, (unsigned long)strlen(cast_const_char text_t->text));
 #endif
 	sl = strlen(cast_const_char ptr);
 	if (sl > MAXINT - sizeof(struct g_object_text)) overalloc();
@@ -300,7 +291,7 @@ static void split_line_object(struct g_part *p, struct g_object *text_go, unsign
 	nt2:
 	if (p->line) for (n = 0; n < p->line->n_entries; n++) if (p->line->entries[n] == text_go) goto found;
 	if (!text_t || text_t != p->text) {
-		internal("split_line_object: bad wrap");
+		internal_error("split_line_object: bad wrap");
 		return;
 	}
 	if (0) {
@@ -459,13 +450,13 @@ int is_in_area(struct map_area *a, int x, int y)
 			return a->ncoords >= 3 && (a->coords[0]-x)*(a->coords[0]-x)+(a->coords[1]-y)*(a->coords[1]-y) <= a->coords[2]*a->coords[2];
 		case SHAPE_POLY:
 			over = 0;
-			if (a->ncoords >= 4) for (i = 0; i + 1 < a->ncoords; i += 2) {
+			if (a->ncoords >= 4) for (i = 0; a->ncoords - i > 1; i += 2) {
 				int x1, x2, y1, y2;
 				x1 = a->coords[i];
 				y1 = a->coords[i + 1];
 				x2 = a->coords[0];
 				y2 = a->coords[1];
-				if (i + 3 < a->ncoords) {
+				if (a->ncoords - i > 3) {
 					x2 = a->coords[i + 2];
 					y2 = a->coords[i + 3];
 				}
@@ -482,7 +473,7 @@ int is_in_area(struct map_area *a, int x, int y)
 			}
 			return over & 1;
 		default:
-			internal("is_in_area: bad shape: %d", a->shape);
+			internal_error("is_in_area: bad shape: %d", a->shape);
 	}
 	return 0;
 }
@@ -519,12 +510,14 @@ static void do_image(struct g_part *p, struct image_description *im)
 			struct memory_list *ml;
 			struct menu_item *menu;
 			struct cache_entry *ce = af->rq->ce;
-			unsigned char *start, *end;
+			unsigned char *start;
+			size_t len;
 			int i;
 			struct image_map *map;
-			if (get_file(af->rq, &start, &end)) goto ft;
-			if (start == end) goto ft;
-			if (get_image_map(ce->head, start, end, tag, &menu, &ml, format_.href_base, format_.target_base, 0, 0, 0, 1)) goto ft;
+			if (get_file(af->rq, &start, &len)) goto ft;
+			if (!len) goto ft;
+			if (len > MAXINT) len = MAXINT;
+			if (get_image_map(ce->head, start, start + len, tag, &menu, &ml, format_.href_base, format_.target_base, 0, 0, 0, 1)) goto ft;
 			map = mem_alloc(sizeof(struct image_map));
 			map->n_areas = 0;
 			for (i = 0; menu[i].text; i++) {
@@ -704,7 +697,7 @@ static void *g_html_special(void *p_, int c, ...)
 			break;
 		default:
 			va_end(l);
-			internal("html_special: unknown code %d", c);
+			internal_error("html_special: unknown code %d", c);
 	}
 	return NULL;
 }
@@ -756,7 +749,7 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 	xstrcmp(format_.link, last_link) || xstrcmp(format_.target, last_target) ||
 	    xstrcmp(format_.image, last_image) || format_.form != last_form
 	    || ((format_.js_event || last_js_event) && compare_js_event_spec(format_.js_event, last_js_event))	) {
-			/*if (!html_format_changed) internal("html_format_changed not set");*/
+			/*if (!html_format_changed) internal_error("html_format_changed not set");*/
 			flush_pending_text_to_line(p);
 			if (xstrcmp(cached_font_face, format_.fontface) || cached_font_face == to_je_ale_prasarna) {
 				if (cached_font_face && cached_font_face != to_je_ale_prasarna) mem_free(cached_font_face);
@@ -848,7 +841,7 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 	    && ((!format_.js_event && !last_js_event) || !compare_js_event_spec(format_.js_event, last_js_event))) {
 		if (!p->data) goto back_link;
 		if (!p->data->nlinks) {
-			internal("no link");
+			internal_error("no link");
 			goto back_link;
 		}
 		link = &p->data->links[p->data->nlinks - 1];
@@ -896,7 +889,7 @@ static void g_put_chars(void *p_, unsigned char *s, int l)
 			link->where = stracpy(last_link);
 			link->target = stracpy(last_target);
 		} else {
-			link->type = last_form->type == FC_TEXT || last_form->type == FC_PASSWORD || last_form->type == FC_FILE ? L_FIELD : last_form->type == FC_TEXTAREA ? L_AREA : last_form->type == FC_CHECKBOX || last_form->type == FC_RADIO ? L_CHECKBOX : last_form->type == FC_SELECT ? L_SELECT : L_BUTTON;
+			link->type = last_form->type == FC_TEXT || last_form->type == FC_PASSWORD || last_form->type == FC_FILE_UPLOAD ? L_FIELD : last_form->type == FC_TEXTAREA ? L_AREA : last_form->type == FC_CHECKBOX || last_form->type == FC_RADIO ? L_CHECKBOX : last_form->type == FC_SELECT ? L_SELECT : L_BUTTON;
 			link->form = last_form;
 			link->target = stracpy(last_form->target);
 		}
@@ -990,7 +983,7 @@ struct g_part *g_format_html_part(unsigned char *start, unsigned char *end, int
 	while (&html_top != e) {
 		kill_html_stack_item(&html_top);
 		if (!&html_top || (void *)&html_top == (void *)&html_stack) {
-			internal("html stack trashed");
+			internal_error("html stack trashed");
 			break;
 		}
 	}
