@@ -131,8 +131,6 @@ static int auth_window(struct object_request *rq, unsigned char *realm)
 		unsigned char *h = get_proxy_string(rq->url);
 		if (!h) h = cast_uchar "";
 		host = display_host(term, h);
-		mem_free(h);
-		host = stracpy(host);
 	} else {
 		unsigned char *h = get_host_name(rq->url);
 		if (!h) return -1;
@@ -197,7 +195,7 @@ struct cert_dialog {
 
 static void cert_action(struct object_request *rq, int yes)
 {
-	if (yes) {
+	if (yes > 0) {
 		rq->hold = 0;
 		change_connection(&rq->stat, NULL, PRI_CANCEL);
 		load_url(rq->url, rq->prev_url, &rq->stat, rq->pri, NC_CACHE, 0, 0, 0);
@@ -214,7 +212,16 @@ static void cert_forall(struct cert_dialog *cs, int yes)
 {
 	struct object_request *rq;
 	struct list_head *lrq;
-	if (yes) add_blacklist_entry(cs->host, cs->bl);
+	if (yes > 0) {
+		add_blacklist_entry(cs->host, cs->bl);
+		del_blacklist_entry(cs->host, BL_AVOID_INSECURE);
+	}
+	if (yes < 0) {
+		add_blacklist_entry(cs->host, BL_AVOID_INSECURE);
+		del_blacklist_entry(cs->host, BL_IGNORE_CERTIFICATE);
+		del_blacklist_entry(cs->host, BL_IGNORE_DOWNGRADE);
+		del_blacklist_entry(cs->host, BL_IGNORE_CIPHER);
+	}
 	foreach(struct object_request, rq, lrq, requests) if (rq->term == cs->term && rq->hold == HOLD_CERT && rq->stat.state == cs->state) {
 		unsigned char *host = get_host_name(rq->url);
 		if (!strcmp(cast_const_char host, cast_const_char cs->host)) cert_action(rq, yes);
@@ -232,6 +239,11 @@ static void cert_no(void *data)
 	cert_forall((struct cert_dialog *)data, 0);
 }
 
+static void cert_never(void *data)
+{
+	cert_forall((struct cert_dialog *)data, -1);
+}
+
 static int cert_compare(void *data1, void *data2)
 {
 	struct cert_dialog *cs1 = (struct cert_dialog *)data1;
@@ -247,6 +259,10 @@ static int cert_window(struct object_request *rq)
 	struct memory_list *ml;
 	if (!(term = find_terminal(rq->term))) return -1;
 	h = get_host_name(rq->url);
+	if (get_blacklist_flags(h) & BL_AVOID_INSECURE) {
+		mem_free(h);
+		return -1;
+	}
 	cs = mem_alloc(sizeof(struct cert_dialog));
 	cs->term = rq->term;
 	cs->host = h;
@@ -272,9 +288,8 @@ static int cert_window(struct object_request *rq)
 	}
 	msg_box(term, ml,
 		title,
-		AL_CENTER | AL_EXTD_TEXT, TEXT_(T_THE_SERVER_), host,
-		text,
-		NULL, cs, 2, TEXT_(T_NO), cert_no, B_ESC, TEXT_(T_YES), cert_yes, B_ENTER);
+		AL_CENTER, TEXT_(T_THE_SERVER_), host, text, MSG_BOX_END,
+		(void *)cs, 3, TEXT_(T_NO), cert_no, B_ESC, TEXT_(T_YES), cert_yes, B_ENTER, TEXT_(T_NEVER), cert_never, 0);
 	return 0;
 }
 
@@ -341,14 +356,13 @@ static void objreq_end(struct status *stat, void *data)
 		if (stat->ce && rq->state == O_WAITING && stat->ce->redirect) {
 			if (rq->redirect_cnt++ < MAX_REDIRECTS) {
 				int cache, allow_flags;
-				unsigned char *u, *p, *pos;
+				unsigned char *u, *pos;
 				change_connection(stat, NULL, PRI_CANCEL);
 				u = join_urls(rq->url, stat->ce->redirect);
 				if ((pos = extract_position(u))) {
 					if (rq->goto_position) mem_free(rq->goto_position);
 					rq->goto_position = pos;
 				}
-				if (!http_options.bug_302_redirect && !stat->ce->redirect_get && (p = cast_uchar strchr(cast_const_char u, POST_CHAR))) add_to_strn(&u, p);
 				cache = rq->cache;
 				if (cache < NC_RELOAD && (!strcmp(cast_const_char u, cast_const_char rq->url) || !strcmp(cast_const_char u, cast_const_char rq->orig_url) || rq->redirect_cnt >= MAX_CACHED_REDIRECTS)) cache = NC_RELOAD;
 				allow_flags = get_allow_flags(rq->url);
@@ -460,7 +474,7 @@ void release_object(struct object_request **rqq)
 void detach_object_connection(struct object_request *rq, off_t pos)
 {
 	if (rq->state == O_WAITING || rq->state == O_FAILED) {
-		internal("detach_object_connection: no data received");
+		internal_error("detach_object_connection: no data received");
 		return;
 	}
 	if (rq->refcount == 1) {
