@@ -5,6 +5,7 @@
 
 #include "links.h"
 
+struct list_head sessions = {&sessions, &sessions};
 struct list_head downloads = {&downloads, &downloads};
 
 /* prototypes */
@@ -24,7 +25,20 @@ int are_there_downloads(void)
 	return 0;
 }
 
-struct list_head sessions = {&sessions, &sessions};
+unsigned char get_session_attribute(struct session *ses, int reverse)
+{
+	if (!ses->term->spec->col) {
+		if (!reverse)
+			return COLOR_TITLE;
+		else
+			return COLOR_STATUS;
+	} else {
+		if (!reverse)
+			return get_attribute(ses->ds.t_text_color, ses->ds.t_background_color);
+		else
+			return get_attribute(ses->ds.t_background_color, ses->ds.t_text_color);
+	}
+}
 
 struct s_msg_dsc {
 	int n;
@@ -46,7 +60,6 @@ static_const struct s_msg_dsc msg_dsc[] = {
 
 	{S__OK,			TEXT_(T_OK)},
 	{S_INTERRUPTED,		TEXT_(T_INTERRUPTED)},
-	{S_EXCEPT,		TEXT_(T_SOCKET_EXCEPTION)},
 	{S_INTERNAL,		TEXT_(T_INTERNAL_ERROR)},
 	{S_OUT_OF_MEM,		TEXT_(T_OUT_OF_MEMORY)},
 	{S_NO_DNS,		TEXT_(T_HOST_NOT_FOUND)},
@@ -98,6 +111,7 @@ static_const struct s_msg_dsc msg_dsc[] = {
 	{S_WAIT_REDIR,		TEXT_(T_WAITING_FOR_REDIRECT_CONFIRMATION)},
 	{0,			NULL}
 };
+
 struct strerror_val {
 	list_entry_1st
 #ifdef REORDER_LIST_ENTRIES
@@ -118,14 +132,14 @@ int get_error_from_errno(int errn)
 {
 	if (errn > 0 && (errn < -S__OK || errn > -S_MAX))
 		return -errn;
-#ifdef BEOS
+#if defined(BEOS) || defined(HAIKU)
 	if (-errn > 0 && (-errn < -S__OK || -errn > -S_MAX))
 		return errn;
 #endif
 	return S_UNKNOWN_ERROR;
 }
 
-unsigned char *get_err_msg(int state)
+unsigned char *get_err_msg(int state, struct terminal *term) /* term may be NULL */
 {
 	unsigned char *e;
 	size_t sl;
@@ -138,17 +152,25 @@ unsigned char *get_err_msg(int state)
 		unk:
 		return TEXT_(T_UNKNOWN_ERROR);
 	}
-#ifdef BEOS
-	if ((e = cast_uchar strerror(state)) && *e && !strstr(cast_const_char e, "No Error")) goto have_error;
+#if defined(BEOS) || defined(HAIKU)
+	e = cast_uchar strerror_alloc(state, term);
+	if (*e && !strstr(cast_const_char e, "No Error")) goto have_error;
+	mem_free(e);
 #endif
-	if ((e = cast_uchar strerror(-state)) && *e) goto have_error;
+	e = cast_uchar strerror_alloc(-state, term);
+	if (*e) goto have_error;
+	mem_free(e);
 	goto unk;
 have_error:
-	foreach(struct strerror_val, s, ls, strerror_buf) if (!strcmp(cast_const_char s->msg, cast_const_char e)) return s->msg;
+	foreach(struct strerror_val, s, ls, strerror_buf) if (!strcmp(cast_const_char s->msg, cast_const_char e)) {
+		mem_free(e);
+		return s->msg;
+	}
 	sl = strlen(cast_const_char e);
 	if (sl > MAXINT - sizeof(struct strerror_val)) overalloc();
 	s = mem_alloc(sizeof(struct strerror_val) + sl);
 	strcpy(cast_char s->msg, cast_const_char e);
+	mem_free(e);
 	add_to_list(strerror_buf, s);
 	return s->msg;
 }
@@ -182,22 +204,31 @@ static unsigned char *get_stat_msg(struct status *stat, struct terminal *term)
 		unsigned char *m = init_str();
 		int l = 0;
 		add_to_str(&m, &l, get_text_translation(TEXT_(T_RECEIVED), term));
-		add_to_str(&m, &l, cast_uchar " ");
+		add_chr_to_str(&m, &l, ' ');
 		add_xnum_to_str(&m, &l, stat->prg->pos);
-		if (stat->prg->size >= 0)
-			add_to_str(&m, &l, cast_uchar " "), add_to_str(&m, &l, get_text_translation(TEXT_(T_OF), term)), add_to_str(&m, &l, cast_uchar " "), add_xnum_to_str(&m, &l, stat->prg->size);
+		if (stat->prg->size >= 0) {
+			add_chr_to_str(&m, &l, ' ');
+			add_to_str(&m, &l, get_text_translation(TEXT_(T_OF), term));
+			add_chr_to_str(&m, &l, ' ');
+			add_xnum_to_str(&m, &l, stat->prg->size);
+		}
 		add_to_str(&m, &l, cast_uchar ", ");
-		if (stat->prg->elapsed >= CURRENT_SPD_AFTER * SPD_DISP_TIME)
-			add_to_str(&m, &l, get_text_translation(TEXT_(T_AVG), term)), add_to_str(&m, &l, cast_uchar " ");
+		if (stat->prg->elapsed >= CURRENT_SPD_AFTER * SPD_DISP_TIME) {
+			add_to_str(&m, &l, get_text_translation(TEXT_(T_AVG), term));
+			add_chr_to_str(&m, &l, ' ');
+		}
 		add_xnum_to_str(&m, &l, stat->prg->loaded * 10 / (stat->prg->elapsed / 100));
 		add_to_str(&m, &l, cast_uchar "/s");
-		if (stat->prg->elapsed >= CURRENT_SPD_AFTER * SPD_DISP_TIME)
-			add_to_str(&m, &l, cast_uchar ", "), add_to_str(&m, &l, get_text_translation(TEXT_(T_CUR), term)), add_to_str(&m, &l, cast_uchar " "),
-			add_xnum_to_str(&m, &l, stat->prg->cur_loaded / (CURRENT_SPD_SEC * SPD_DISP_TIME / 1000)),
+		if (stat->prg->elapsed >= CURRENT_SPD_AFTER * SPD_DISP_TIME) {
+			add_to_str(&m, &l, cast_uchar ", ");
+			add_to_str(&m, &l, get_text_translation(TEXT_(T_CUR), term));
+			add_chr_to_str(&m, &l, ' ');
+			add_xnum_to_str(&m, &l, stat->prg->cur_loaded / (CURRENT_SPD_SEC * SPD_DISP_TIME / 1000));
 			add_to_str(&m, &l, cast_uchar "/s");
+		}
 		return m;
 	}
-	return stracpy(get_text_translation(get_err_msg(stat->state), term));
+	return stracpy(get_text_translation(get_err_msg(stat->state, term), term));
 }
 
 void change_screen_status(struct session *ses)
@@ -247,15 +278,14 @@ static void x_print_screen_status(struct terminal *term, void *ses_)
 {
 	struct session *ses = (struct session *)ses_;
 	if (!F) {
-		unsigned char color = get_attribute(ses->ds.t_text_color, ses->ds.t_background_color);
-		if (!term->spec->col) color = COLOR_TITLE;
+		unsigned char color = get_session_attribute(ses, proxies.only_proxies);
 		fill_area(term, 0, term->y - 1, term->x, 1, ' ', color);
 		if (ses->st) print_text(term, 0, term->y - 1, (int)strlen(cast_const_char ses->st), ses->st, COLOR_STATUS);
 #ifdef G
 	} else {
 		int l = 0;
 		if (ses->st) g_print_text(term->dev, 0, term->y - G_BFU_FONT_SIZE, bfu_style_wb_mono, ses->st, &l);
-		drv->fill_area(term->dev, l, term->y - G_BFU_FONT_SIZE, term->x, term->y, bfu_bg_color);
+		drv->fill_area(term->dev, l, term->y - G_BFU_FONT_SIZE, term->x, term->y, !proxies.only_proxies ? bfu_bg_color : bfu_fg_color);
 #endif
 	}
 }
@@ -264,7 +294,8 @@ static void x_print_screen_title(struct terminal *term, void *ses_)
 {
 	struct session *ses = (struct session *)ses_;
 	unsigned char *m;
-	unsigned char color = get_attribute(ses->ds.t_text_color, ses->ds.t_background_color);
+	unsigned char color = get_session_attribute(ses, proxies.only_proxies);
+	if (!term->spec->col) color = COLOR_TITLE;
 	fill_area(term, 0, 0, term->x, 1, ' ', color);
 	if ((m = print_current_title(ses))) {
 		int p = term->x - 1 - cp_len(term_charset(ses->term), m);
@@ -279,9 +310,6 @@ static void print_only_screen_status(struct session *ses)
 {
 #ifdef G
 	if (F) {
-		/*debug("%s - %s", ses->st_old, ses->st);
-		debug("clip: %d.%d , %d.%d", ses->term->dev->clip.x1, ses->term->dev->clip.y1, ses->term->dev->clip.x2, ses->term->dev->clip.y2);
-		debug("size: %d.%d , %d.%d", ses->term->dev->size.x1, ses->term->dev->size.y1, ses->term->dev->size.x2, ses->term->dev->size.y2);*/
 		if (ses->st_old) {
 			if (ses->st && !strcmp(cast_const_char ses->st, cast_const_char ses->st_old)) return;
 			mem_free(ses->st_old);
@@ -302,7 +330,10 @@ void print_screen_status(struct session *ses)
 		draw_to_window(ses->win, x_print_screen_title, ses);
 
 	m = stracpy(cast_uchar "Links");
-	if (ses->screen && ses->screen->f_data && ses->screen->f_data->title && ses->screen->f_data->title[0]) add_to_strn(&m, cast_uchar " - "), add_to_strn(&m, ses->screen->f_data->title);
+	if (ses->screen && ses->screen->f_data && ses->screen->f_data->title && ses->screen->f_data->title[0]) {
+		add_to_strn(&m, cast_uchar " - ");
+		add_to_strn(&m, ses->screen->f_data->title);
+	}
 	set_terminal_title(ses->term, m);
 	/*mem_free(m); -- set_terminal_title frees it */
 
@@ -322,7 +353,7 @@ void print_progress(struct session *ses, unsigned char *msg)
 
 void print_error_dialog(struct session *ses, struct status *stat, unsigned char *url)
 {
-	unsigned char *t = get_err_msg(stat->state);
+	unsigned char *t = get_err_msg(stat->state, ses->term);
 	unsigned char *u = display_url(ses->term, url, 1);
 	msg_box(ses->term, getml(u, NULL), TEXT_(T_ERROR), AL_CENTER, TEXT_(T_ERROR_LOADING), cast_uchar " ", u, cast_uchar ":\n\n", t, MSG_BOX_END, (void *)ses, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC /*, get_text_translation("Retry"), NULL, 0 !!! FIXME: retry */);
 }
@@ -548,34 +579,43 @@ void download_window_function(struct dialog_data *dlg)
 		m = init_str();
 		t = 1;
 		add_to_str(&m, &l, get_text_translation(TEXT_(T_RECEIVED), term));
-		add_to_str(&m, &l, cast_uchar " ");
+		add_chr_to_str(&m, &l, ' ');
 		add_xnum_to_str(&m, &l, stat->prg->pos);
-		if (stat->prg->size >= 0)
-			add_to_str(&m, &l, cast_uchar " "), add_to_str(&m, &l, get_text_translation(TEXT_(T_OF),term)), add_to_str(&m, &l, cast_uchar " "), add_xnum_to_str(&m, &l, stat->prg->size), add_to_str(&m, &l, cast_uchar " ");
+		if (stat->prg->size >= 0) {
+			add_chr_to_str(&m, &l, ' ');
+			add_to_str(&m, &l, get_text_translation(TEXT_(T_OF),term));
+			add_chr_to_str(&m, &l, ' ');
+			add_xnum_to_str(&m, &l, stat->prg->size);
+			add_chr_to_str(&m, &l, ' ');
+		}
 		add_to_str(&m, &l, cast_uchar "\n");
 		if (stat->prg->elapsed >= CURRENT_SPD_AFTER * SPD_DISP_TIME)
 			add_to_str(&m, &l, get_text_translation(TEXT_(T_AVERAGE_SPEED), term));
-		else add_to_str(&m, &l, get_text_translation(TEXT_(T_SPEED), term));
-		add_to_str(&m, &l, cast_uchar " ");
+		else
+			add_to_str(&m, &l, get_text_translation(TEXT_(T_SPEED), term));
+		add_chr_to_str(&m, &l, ' ');
 		add_xnum_to_str(&m, &l, (longlong)stat->prg->loaded * 10 / (stat->prg->elapsed / 100));
 		add_to_str(&m, &l, cast_uchar "/s");
-		if (stat->prg->elapsed >= CURRENT_SPD_AFTER * SPD_DISP_TIME)
-			add_to_str(&m, &l, cast_uchar ", "), add_to_str(&m, &l, get_text_translation(TEXT_(T_CURRENT_SPEED), term)), add_to_str(&m, &l, cast_uchar " "),
-			add_xnum_to_str(&m, &l, stat->prg->cur_loaded / (CURRENT_SPD_SEC * SPD_DISP_TIME / 1000)),
+		if (stat->prg->elapsed >= CURRENT_SPD_AFTER * SPD_DISP_TIME) {
+			add_to_str(&m, &l, cast_uchar ", ");
+			add_to_str(&m, &l, get_text_translation(TEXT_(T_CURRENT_SPEED), term));
+			add_chr_to_str(&m, &l, ' ');
+			add_xnum_to_str(&m, &l, stat->prg->cur_loaded / (CURRENT_SPD_SEC * SPD_DISP_TIME / 1000));
 			add_to_str(&m, &l, cast_uchar "/s");
+		}
 		add_to_str(&m, &l, cast_uchar "\n");
 		add_to_str(&m, &l, get_text_translation(TEXT_(T_ELAPSED_TIME), term));
-		add_to_str(&m, &l, cast_uchar " ");
+		add_chr_to_str(&m, &l, ' ');
 		add_time_to_str(&m, &l, stat->prg->elapsed / 1000);
 		if (stat->prg->size >= 0 && stat->prg->loaded > 0) {
 			add_to_str(&m, &l, cast_uchar ", ");
 			add_to_str(&m, &l, get_text_translation(TEXT_(T_ESTIMATED_TIME), term));
-			add_to_str(&m, &l, cast_uchar " ");
+			add_chr_to_str(&m, &l, ' ');
 			/*add_time_to_str(&m, &l, stat->prg->elapsed / 1000 * stat->prg->size / stat->prg->loaded * 1000 - stat->prg->elapsed);*/
 			/*add_time_to_str(&m, &l, (stat->prg->size - stat->prg->pos) / ((longlong)stat->prg->loaded * 10 / (stat->prg->elapsed / 100)));*/
 			add_time_to_str(&m, &l, (uttime)((stat->prg->size - stat->prg->pos) / ((double)stat->prg->loaded * 1000 / stat->prg->elapsed)));
 		}
-	} else m = stracpy(get_text_translation(get_err_msg(stat->state), term));
+	} else m = stracpy(get_text_translation(get_err_msg(stat->state, term), term));
 	show_percentage = t && test_percentage(stat);
 	u = display_url(term, down->url, 1);
 	max_text_width(term, u, &max, AL_LEFT);
@@ -811,10 +851,11 @@ time_t parse_http_date(unsigned char *date)	/* this functions is bad !!! */
 
 static void download_file_error(struct download *down, int err)
 {
-	if (get_download_ses(down)) {
-		unsigned char *emsg = stracpy(err ? cast_uchar strerror(err) : cast_uchar "Zero returned");
+	struct session *ses = get_download_ses(down);
+	if (ses) {
+		unsigned char *emsg = err ? strerror_alloc(err, ses->term) : stracpy(cast_uchar "Zero returned");
 		unsigned char *msg = stracpy(down->file);
-		msg_box(get_download_ses(down)->term, getml(msg, emsg, NULL), TEXT_(T_DOWNLOAD_ERROR), AL_CENTER, TEXT_(T_COULD_NOT_WRITE_TO_FILE), cast_uchar " ", msg, cast_uchar ": ", emsg, MSG_BOX_END, NULL, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC);
+		msg_box(ses->term, getml(msg, emsg, NULL), TEXT_(T_DOWNLOAD_ERROR), AL_CENTER, TEXT_(T_COULD_NOT_WRITE_TO_FILE), cast_uchar " ", msg, cast_uchar ": ", emsg, MSG_BOX_END, NULL, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC);
 	}
 }
 
@@ -953,18 +994,19 @@ have_frag:
 	if (stat->state < 0) {
 		if (down->decompress) {
 			struct session *ses = get_download_ses(down);
-			unsigned char *start, *end;
+			unsigned char *start;
+			size_t len;
 			int err;
-			get_file_by_term(ses ? ses->term : NULL, ce, &start, &end, &err);
+			get_file_by_term(ses ? ses->term : NULL, ce, &start, &len, &err);
 			if (err) goto det_abt;
-			if (download_prealloc(down, end - start))
+			if (download_prealloc(down, len))
 				goto det_abt;
-			while (down->last_pos < end - start) {
-				if (download_write(down, start + down->last_pos, end - start - down->last_pos)) goto det_abt;
+			while (down->last_pos + (tcount)0 < len) {
+				if (download_write(down, start + down->last_pos, len - down->last_pos)) goto det_abt;
 			}
 		}
 		if (stat->state != S__OK) {
-			unsigned char *t = get_err_msg(stat->state);
+			unsigned char *t = get_err_msg(stat->state, get_download_ses(down)->term);
 			unsigned char *tt = display_url(get_download_ses(down)->term, down->url, 1);
 			msg_box(get_download_ses(down)->term, getml(tt, NULL), TEXT_(T_DOWNLOAD_ERROR), AL_CENTER, TEXT_(T_ERROR_DOWNLOADING), cast_uchar " ", tt, cast_uchar ":\n\n", t, MSG_BOX_END, (void *)get_download_ses(down), 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC /*, TEXT_(T_RETRY), NULL, 0 !!! FIXME: retry */);
 		} else {
@@ -1072,7 +1114,7 @@ int create_download_file(struct session *ses, unsigned char *cwd, unsigned char
 		}
 		if (!ses) goto x;
 		msg = stracpy(file);
-		msge = stracpy(cast_uchar strerror(errn));
+		msge = strerror_alloc(errn, ses->term);
 		msg_box(ses->term, getml(msg, msge, NULL), TEXT_(T_DOWNLOAD_ERROR), AL_CENTER, TEXT_(T_COULD_NOT_CREATE_FILE), cast_uchar " ", msg, cast_uchar ": ", msge, MSG_BOX_END, NULL, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC);
 		goto x;
 	}
@@ -1108,7 +1150,7 @@ static int create_or_append_download_file(struct session *ses, unsigned char *cw
 
 		EINTRLOOP(ls, lseek(*hp, 0, SEEK_END));
 		if (ls == (off_t)-1) {
-			unsigned char *emsg = stracpy(cast_uchar strerror(errno));
+			unsigned char *emsg = strerror_alloc(errno, ses->term);
 			unsigned char *msg = stracpy(*xl_file);
 			msg_box(ses->term, getml(msg, emsg, NULL), TEXT_(T_DOWNLOAD_ERROR), AL_CENTER, TEXT_(T_ERROR_CALLING_LSEEK_ON_FILE), cast_uchar " ", msg, cast_uchar ": ", emsg, MSG_BOX_END, NULL, 1, TEXT_(T_CANCEL), msg_box_null, B_ENTER | B_ESC);
 			goto err_close;
@@ -1312,24 +1354,26 @@ static struct f_data *format_html(struct f_data_c *fd, struct object_request *rq
 	f->time_to_get = -get_time();
 	clone_object(rq, &f->rq);
 	if (f->rq->ce) {
-		unsigned char *start; unsigned char *end;
+		unsigned char *start;
+		size_t len;
 		int stl = -1;
 		struct additional_file *af;
 		struct list_head *laf;
 
 		if (fd->af) foreach(struct additional_file, af, laf, fd->af->af) if (af->need_reparse > 0) af->need_reparse = 0;
 
-		get_file(rq, &start, &end);
-		if (jsint_get_source(fd, &start, &end)) f->uncacheable = 1;
+		get_file(rq, &start, &len);
+		if (jsint_get_source(fd, &start, &len)) f->uncacheable = 1;
+		if (len > MAXINT) len = MAXINT;
 		if (opt->plain == 2) {
 			start = init_str();
 			stl = 0;
 			add_to_str(&start, &stl, cast_uchar "<img src=\"");
 			add_to_str(&start, &stl, f->rq->ce->url);
 			add_to_str(&start, &stl, cast_uchar "\">");
-			end = start + stl;
+			len = stl;
 		}
-		really_format_html(f->rq->ce, start, end, f, fd->ses ? fd != fd->ses->screen : 0);
+		really_format_html(f->rq->ce, start, start + len, f, fd->ses ? fd != fd->ses->screen : 0);
 		if (stl != -1) mem_free(start);
 		f->use_tag = f->rq->ce->count;
 		if (f->af) foreach(struct additional_file, af, laf, f->af->af) {
@@ -1371,7 +1415,7 @@ unsigned long formatted_info(int type)
 				count_frames(ses->screen, &i);
 			return i;
 		default:
-			internal("formatted_info: bad request");
+			internal_error("formatted_info: bad request");
 	}
 	return 0;
 }
@@ -1438,7 +1482,7 @@ static void detach_f_data(struct f_data **ff)
 	}
 }
 
-static int shrink_format_cache(int u)
+int shrink_format_cache(int u)
 {
 	static int sc = 0;
 	int scc;
@@ -1552,7 +1596,7 @@ struct f_data *cached_format_html(struct f_data_c *fd, struct object_request *rq
 		}
 	}) {};
 	if (ses) {
-		if (report_status || !fd->f_data || fd->f_data->time_to_get >= DISPLAY_FORMATTING_STATUS)
+		if (report_status || !fd->f_data || fd->f_data->time_to_get >= DISPLAY_FORMATTING_STATUS || (rq->ce && rq->ce->length >= 1000000))
 			print_progress(ses, TEXT_(T_FORMATTING_DOCUMENT));
 	}
 	detach_f_data(&fd->f_data);
@@ -1669,12 +1713,12 @@ static void html_interpret(struct f_data_c *fd, int report_status)
 #ifdef JS
 	o.js_enable = js_enable;
 #else
-	o.js_enable=0;
+	o.js_enable = 0;
 #endif
 #ifdef G
-	o.bfu_aspect=bfu_aspect;
+	o.gamma_stamp = gamma_stamp;
 #else
-	o.bfu_aspect=0;
+	o.gamma_stamp = 0;
 #endif
 	o.plain = fd->vs->plain;
 	if (o.plain == 1 && !o.break_long_lines) {
@@ -1960,7 +2004,24 @@ int f_data_c_allow_flags(struct f_data_c *fd)
 	return 0;
 }
 
-static int plain_type(struct session *ses, struct object_request *rq, unsigned char **p)
+static int is_forced_download(struct object_request *rq)
+{
+	struct cache_entry *ce;
+	unsigned char *cd;
+	if (!rq || !(ce = rq->ce))
+		return 0;
+	if ((cd = parse_http_header(ce->head, cast_uchar "Content-Disposition", NULL))) {
+		int r;
+		unsigned char *s;
+		if ((s = cast_uchar strchr(cast_const_char cd, ';'))) *s = 0;
+		r = !casestrcmp(cd, cast_uchar "attachment");
+		mem_free(cd);
+		return r;
+	}
+	return 0;
+}
+
+static int plain_type(struct object_request *rq, unsigned char **p)
 {
 	struct cache_entry *ce;
 	unsigned char *ct;
@@ -2026,7 +2087,7 @@ void fd_loaded(struct object_request *rq, void *fd_)
 	}
 	if (fd->parsed_done && f_need_reparse(fd->f_data)) fd->parsed_done = 0;
 	if (fd->vs->plain == -1 && rq->state != O_WAITING) {
-		fd->vs->plain = plain_type(fd->ses, fd->rq, NULL);
+		fd->vs->plain = plain_type(fd->rq, NULL);
 	}
 	if (fd->rq->state < 0 && (f_is_finished(fd->f_data) || !fd->f_data)) {
 		if (!fd->parsed_done) {
@@ -2151,7 +2212,7 @@ static struct location *copy_sublocations(struct session *ses, struct location *
 		add_to_list_end(d->subframes, dl);
 		dl->parent = d;
 		z = copy_sublocations(ses, dl, sl, x);
-		if (z && y) internal("copy_sublocations: crossed references");
+		if (z && y) internal_error("copy_sublocations: crossed references");
 		if (z) y = z;
 	}
 	return y;
@@ -2162,7 +2223,7 @@ static struct location *copy_location(struct session *ses, struct location *loc)
 	struct location *l2, *l1, *nl;
 	l1 = cur_loc(ses);
 	l2 = alloc_ses_location(ses);
-	if (!(nl = copy_sublocations(ses, l2, l1, loc))) internal("copy_location: sublocation not found");
+	if (!(nl = copy_sublocations(ses, l2, l1, loc))) internal_error("copy_location: sublocation not found");
 	return nl;
 }
 
@@ -2264,7 +2325,8 @@ static void ses_go_forward(struct session *ses, int plain, int refresh)
 	fd->rq->upcall = fd_loaded;
 	fd->rq->data = fd;
 	fd->rq->upcall(fd->rq, fd);
-	draw_formatted(ses);
+	if (!list_empty(ses->screen->subframes))
+		draw_formatted(ses);
 }
 
 static void ses_go_backward(struct session *ses)
@@ -2275,13 +2337,13 @@ static void ses_go_backward(struct session *ses)
 	if (ses->search_word) mem_free(ses->search_word), ses->search_word = NULL;
 	if (ses->default_status){mem_free(ses->default_status);ses->default_status=NULL;}	/* smazeme default status, aby neopruzoval na jinych strankach */
 	reinit_f_data_c(ses->screen);
-	if (!ses->wtd_num_steps) internal("ses_go_backward: wtd_num_steps is zero");
+	if (!ses->wtd_num_steps) internal_error("ses_go_backward: wtd_num_steps is zero");
 	if (ses->wtd_num_steps > 0) {
 		n = ses->wtd_num_steps;
 		foreach(struct location, loc, lloc, ses->history) {
 			if (!n--) goto have_back_loc;
 		}
-		internal("ses_go_backward: session history disappeared");
+		internal_error("ses_go_backward: session history disappeared");
 		return;
 		have_back_loc:
 		for (n = 0; n < ses->wtd_num_steps; n++) {
@@ -2294,7 +2356,7 @@ static void ses_go_backward(struct session *ses)
 		foreach(struct location, loc, lloc, ses->forward_history) {
 			if (!++n) goto have_fwd_loc;
 		}
-		internal("ses_go_backward: session forward history disappeared");
+		internal_error("ses_go_backward: session forward history disappeared");
 		return;
 		have_fwd_loc:
 		for (n = 0; n < -ses->wtd_num_steps; n++) {
@@ -2419,10 +2481,13 @@ static int ses_abort_1st_state_loading(struct session *ses)
 static void tp_display(void *ses_)
 {
 	struct session *ses = (struct session *)ses_;
+	int plain = 1;
+	if (plain_type(ses->tq, NULL) == 2)
+		plain = 2;
 	ses_abort_1st_state_loading(ses);
 	ses->rq = ses->tq;
 	ses->tq = NULL;
-	ses_go_forward(ses, 1, 0);
+	ses_go_forward(ses, plain, 0);
 }
 
 static int direct_download_possible(struct object_request *rq, struct assoc *a)
@@ -2472,7 +2537,7 @@ static int prog_sel_open(struct dialog_data *dlg, struct dialog_item_data *idata
 	struct assoc *a = (struct assoc*)idata->item->udata;
 	struct session *ses = (struct session *)dlg->dlg->udata2;
 
-	if (!a) internal("This should not happen.\n");
+	if (!a) internal_error("This should not happen.\n");
 	ses->tq_prog = stracpy(a->prog), ses->tq_prog_flag_block = a->block, ses->tq_prog_flag_direct = direct_download_possible(ses->tq, a);
 	tp_open(ses);
 	cancel_dialog(dlg,idata);
@@ -2574,9 +2639,10 @@ static void ses_go_to_2nd_state(struct session *ses)
 	struct assoc *a;
 	int n;
 	unsigned char *ct = NULL;
-	int r = plain_type(ses, ses->rq, &ct);
-	if (r == 0 || r == 1 || r == 2 || r == 3) goto go;
-	if (!(a = get_type_assoc(ses->term, ct, &n)) && strlen(cast_const_char ct) >= 4 && !casecmp(ct, cast_uchar "text", 4)) {
+	int r = plain_type(ses->rq, &ct);
+	int force_download = is_forced_download(ses->rq);
+	if (!force_download && (r == 0 || r == 1 || r == 2)) goto go;
+	if (!(a = get_type_assoc(ses->term, ct, &n)) && strlen(cast_const_char ct) >= 4 && !casecmp(ct, cast_uchar "text", 4) && !force_download) {
 		r = 1;
 		goto go;
 	}
@@ -2757,15 +2823,17 @@ static void freeml_void(void *ml_)
 
 static void ses_imgmap(struct session *ses)
 {
-	unsigned char *start, *end;
+	unsigned char *start;
+	size_t len;
 	struct memory_list *ml;
 	struct menu_item *menu;
 	struct f_data_c *fd;
 	if (ses->rq->state != O_OK && ses->rq->state != O_INCOMPLETE) return;
 	if (!(fd = current_frame(ses)) || !fd->f_data) return;
-	if (get_file(ses->rq, &start, &end)) return;
+	if (get_file(ses->rq, &start, &len)) return;
+	if (len > MAXINT) len = MAXINT;
 	d_opt = &fd->f_data->opt;
-	if (get_image_map(ses->rq->ce->head, start, end, ses->goto_position, &menu, &ml, ses->imgmap_href_base, ses->imgmap_target_base, term_charset(ses->term), ses->ds.assume_cp, ses->ds.hard_assume, 0)) {
+	if (get_image_map(ses->rq->ce->head, start, start + len, ses->goto_position, &menu, &ml, ses->imgmap_href_base, ses->imgmap_target_base, term_charset(ses->term), ses->ds.assume_cp, ses->ds.hard_assume, 0)) {
 		ses_abort_1st_state_loading(ses);
 		return;
 	}
@@ -3061,8 +3129,14 @@ void win_func(struct window *win, struct links_event *ev, int fw)
 			if (ev->ev == EV_KBD || (ev->b & BM_ACT) != B_MOVE || BM_IS_WHEEL(ev->b)) move_session_to_front(ses);
 			send_event(ses, ev);
 			break;
+		case EV_EXTRA:
+			if (ev->x == EV_EXTRA_OPEN_URL) {
+				goto_url(ses, (unsigned char *)ev->b);
+				break;
+			}
+			break;
 		default:
-			error("ERROR: unknown event");
+			break;
 	}
 }
 
