@@ -24,93 +24,6 @@
 
 #include "bits.h"
 
-#if defined(__CYGWIN__) && defined(HAVE_LIBFONTCONFIG)
-#ifdef HAVE_PTHREADS
-#define BACKGROUND_FONT_INIT
-#endif
-#include <fontconfig/fontconfig.h>
-#include <windows.h>
-#include <w32api/shlobj.h>
-static void set_font_path(void)
-{
-	unsigned char *path = cast_uchar "/cygdrive/c/Windows/Fonts";
-#if defined(HAVE_CYGWIN_CONV_PATH)
-	unsigned char win32_path[MAX_PATH];
-	if (SHGetFolderPathA(NULL, CSIDL_FONTS, NULL, 0, cast_char win32_path) == S_OK) {
-		ssize_t l;
-		unsigned char *cyg_path;
-		l = cygwin_conv_path(CCP_WIN_A_TO_POSIX, win32_path, NULL, 0);
-		if (l <= 0)
-			goto do_default;
-		cyg_path = alloca(l);
-		l = cygwin_conv_path(CCP_WIN_A_TO_POSIX, win32_path, cyg_path, l);
-		if (l < 0)
-			goto do_default;
-		path = cyg_path;
-	}
-	do_default:
-#endif
-	FcConfigAppFontAddDir(NULL, path);
-}
-#else
-#define set_font_path()	do { } while (0)
-#endif
-
-static void *do_initialize_fonts(void *ptr)
-{
-#if !defined(GLIB_DEPRECATED_IN_2_36)
-	g_type_init();
-#endif
-#if !LIBRSVG_CHECK_VERSION(2,36,0)
-	rsvg_init();
-#endif
-	set_font_path();
-	return NULL;
-}
-
-
-#ifdef BACKGROUND_FONT_INIT
-
-#include <pthread.h>
-
-static pthread_t font_thread;
-
-void spawn_font_thread(void)
-{
-	int r;
-	if ((r = pthread_create(&font_thread, NULL, do_initialize_fonts, NULL)))
-		fatal_exit("Could not start thread: %s", strerror(r));
-}
-
-static void wait_for_fonts(void)
-{
-	static unsigned char font_initialized = 0;
-	if (!font_initialized) {
-		int r;
-		if ((r = pthread_join(font_thread, NULL)))
-			fatal_exit("pthread_join failed: %s", strerror(r));
-		font_initialized = 1;
-	}
-}
-
-#else
-
-void spawn_font_thread(void)
-{
-}
-
-static void wait_for_fonts(void)
-{
-	static unsigned char font_initialized = 0;
-	if (!font_initialized) {
-		do_initialize_fonts(NULL);
-		font_initialized = 1;
-	}
-}
-
-#endif
-
-
 struct svg_decoder {
 	RsvgHandle *handle;
 #ifdef CURRENTCOLOR_HACK
@@ -123,7 +36,20 @@ void svg_start(struct cached_image *cimg)
 {
 	struct svg_decoder *deco;
 
-	wait_for_fonts();
+	wait_for_fontconfig();
+
+	{
+		static unsigned rsvg_initialized = 0;
+		if (!rsvg_initialized) {
+#if !defined(GLIB_DEPRECATED_IN_2_36)
+			g_type_init();
+#endif
+#if !LIBRSVG_CHECK_VERSION(2,36,0)
+			rsvg_init();
+#endif
+			rsvg_initialized = 1;
+		}
+	}
 
 	deco = mem_alloc(sizeof(struct svg_decoder));
 #ifdef CURRENTCOLOR_HACK
@@ -217,13 +143,17 @@ void svg_finish(struct cached_image *cimg)
 
 	cimg->width = dim.width;
 	cimg->height = dim.height;
+	if (strstr(cast_const_char cimg->url, "/media/math/render/svg/")) {
+		cimg->width = (ssize_t)(cimg->width * 1.36);
+		cimg->height = (ssize_t)(cimg->height * 1.36);
+	}
 	cimg->buffer_bytes_per_pixel = 4;
-	cimg->red_gamma=cimg->green_gamma=cimg->blue_gamma=(float)sRGB_gamma;
-	cimg->strip_optimized=0;
+	cimg->red_gamma = cimg->green_gamma = cimg->blue_gamma = (float)sRGB_gamma;
+	cimg->strip_optimized = 0;
 	if (header_dimensions_known(cimg))
 		goto end;
 
-	surf = cairo_image_surface_create_for_data(cimg->buffer, CAIRO_FORMAT_ARGB32, cimg->width, cimg->height, cimg->width * cimg->buffer_bytes_per_pixel);
+	surf = cairo_image_surface_create_for_data(cimg->buffer, CAIRO_FORMAT_ARGB32, (int)cimg->width, (int)cimg->height, (int)(cimg->width * cimg->buffer_bytes_per_pixel));
 	if (cairo_surface_status(surf))
 		goto end_surface;
 
@@ -240,7 +170,7 @@ void svg_finish(struct cached_image *cimg)
 	restore_stderr(h);
 
 	end_buffer = cimg->buffer + cimg->width * cimg->height * cimg->buffer_bytes_per_pixel;
-	if (htonl(0x12345678L) != 0x12345678L) {
+	if (!big_endian) {
 		for (p = cimg->buffer; p < end_buffer; p += 4) {
 #ifdef t4c
 			t4c t = *(t4c *)p;
@@ -300,7 +230,7 @@ void add_svg_version(unsigned char **s, int *l)
 	add_chr_to_str(s, l, '.');
 	add_num_to_str(s, l, LIBRSVG_MICRO_VERSION);
 #endif
-	add_to_str(s, l, cast_uchar ")");
+	add_chr_to_str(s, l, ')');
 }
 
 #endif
